/* pwcheck_ldap.c -- check passwords using LDAP
 *
 * Author: Clayton Donley <donley@cig.mot.com>
 *         http://www.wwa.com/~donley/
 * Version: 1.01
 *
 * Note: This works by finding a DN that matches an entered UID and
 * binding to the LDAP server using this UID.
 *
 * This version should work with both University of Michigan and Netscape
 * LDAP libraries.  It also gets rid of the requirement for userPassword
 * attribute readability.
 *
 *
 * Minor changes by Simon Loader <simon@surf.org.uk>
 *	Mainly to do with making it slightly more configurable and
 * some memory leaks to do with openLdap 1.1.2 Version
 *  NOT TESTED AGAINST NETSCAPE.
 *
 * Changed by Andy Igoshin <ai@vsu.ru>
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <lber.h>
#include <ldap.h>
#if defined(TAC_LDAP_VERSION3)
#if defined(TAC_LDAP_NS_SSL)
#include <ldap_ssl.h>
#endif /* TAC_LDAP_NS_SSL */
#endif /* TAC_LDAP_VERSION3 */
#include "tac_plus.h"


/* Set These to your Local Environment */

#ifndef MY_LDAP_UIDATTR
#define MY_LDAP_UIDATTR	"uid"
#endif

int pwcheck_ldap(char *userid, char *passwd, char *basedn,
    char *server, char *binddn, char *bindpw,
    char *saslmech, char *saslcred, char *sslpath)
{
    LDAP *ld;
    LDAPMessage *result;
    LDAPMessage *entry;
    char *attrs[2];
    char filter[200];
    char *dn;
    int rc;
#if defined(TAC_LDAP_VERSION3)
    int version = LDAP_VERSION3;
#if defined(TAC_LDAP_SASL)
    struct berval cred;
    struct berval *servcred;
#endif /* TAC_LDAP_SASL */
#endif /* TAC_LDAP_VERSION3 */


    if (strcmp(userid, "") == 0 || strcmp(passwd, "") == 0 ||
	strcmp(basedn, "") == 0 || strcmp(server, "") == 0) {
	report(LOG_ERR, "pwcheck_ldap: userid='%s', passwd='%s', basedn='%s', server='%s'",
	       userid, passwd, basedn, server);
	return 0;
    }
#if defined(TAC_LDAP_SASL)
    if (strcmp(saslmech, "") == 0 || strcmp(saslcred, "") == 0) {
	report(LOG_ERR, "pwcheck_ldap: saslmech='%s', saslcred='%s'",
	       saslmech, saslcred);
	return 0;
    }
#endif /* TAC_LDAP_SASL */
#if defined(TAC_LDAP_NS_SSL)
    if (strcmp(sslpath, "") == 0) {
	report(LOG_ERR, "pwcheck_ldap: sslpath='%s'", sslpath);
	return 0;
    }
#endif /* TAC_LDAP_NS_SSL */

#if defined(TAC_LDAP_VERSION3)
#if defined(TAC_LDAP_NS_SSL)
    if ((ld = ldapssl_client_init(sslpath, NULL)) == NULL) {
	report(LOG_ERR, "pwcheck_ldap: ldapssl_client_init error");
	return 0;
    }
    if ((ld = ldapssl_init(server, LDAPS_PORT, 1)) == NULL) {
	report(LOG_ERR, "pwcheck_ldap: ldapssl_init error");
#endif /* TAC_LDAP_NS_SSL */
#if !defined(TAC_LDAP_NS_SSL) && !defined(TAC_LDAP_SSL)
    if ((ld = ldap_init(server, LDAP_PORT)) == NULL) {
	report(LOG_ERR, "pwcheck_ldap: ldap_init error");
#endif
#else
    if ((ld = ldap_open(server, LDAP_PORT)) == NULL) {
	report(LOG_ERR, "pwcheck_ldap: ldap_open error");
#endif /* TAC_LDAP_VERSION3 */
	return 0; /* could contact it something is wrong */
    }

#if defined(TAC_LDAP_VERSION3)
    ldap_set_option(ld, LDAP_OPT_PROTOCOL_VERSION, &version);
#endif /* TAC_LDAP_VERSION3 */

#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_SASL)
    cred.bv_val = saslcred;
    cred.bv_len = strlen(saslcred);
    if ((rc = ldap_sasl_bind_s(ld, binddn, saslmech,
			       &cred, NULL, NULL, &servcred)) != LDAP_SUCCESS) {
	report(LOG_ERR, "pwcheck_ldap: ldap_sasl_bind_s 1 error: %s", ldap_err2string(rc));
#else
    if ((rc = ldap_simple_bind_s(ld, binddn, bindpw)) != LDAP_SUCCESS) {
	report(LOG_ERR, "pwcheck_ldap: ldap_simple_bind_s 1 error: %s", ldap_err2string(rc));
#endif /* TAC_LDAP_VERSION3 && TAC_LDAP_SASL */
	ldap_unbind(ld);
	return 0;
    }

/* Generate a filter that will return the entry with a matching UID */

    snprintf(filter, sizeof(filter), "(%s=%s)", MY_LDAP_UIDATTR, userid);

/* Just return country... This doesn't actually matter, since we will
   not read the attributes and values, only the DN */

    attrs[0] = "c";
    attrs[1] = NULL;

/* Perform the search... */

#if defined(TAC_LDAP_VERSION3)
    if ((rc = ldap_search_ext_s(ld, basedn, LDAP_SCOPE_SUBTREE, filter, attrs,
			        1, NULL, NULL, NULL, LDAP_NO_LIMIT, &result)) != LDAP_SUCCESS) {
	report(LOG_ERR, "pwcheck_ldap: ldap_search_ext_s error: %s", ldap_err2string(rc));
#else
    if ((rc = ldap_search_s(ld, basedn, LDAP_SCOPE_SUBTREE, filter, attrs,
			    1, &result)) != LDAP_SUCCESS) {
	report(LOG_ERR, "pwcheck_ldap: ldap_search_s error: %s", ldap_err2string(rc));
#endif /* TAC_LDAP_VERSION3 */
	ldap_unbind(ld);
	return 0; /* search filed for some reason */
    }

/* If the entry count is not equal to one, either the UID was not unique or
   there was no match */

    if ((rc = ldap_count_entries(ld, result)) != 1) {
	report(LOG_ERR, "pwcheck_ldap: ldap_count_entries error: rc = %d", rc);
	ldap_msgfree(result);
	ldap_unbind(ld);
	return 0; /* could find uid */
    }

/* Get the first entry */

    if ((entry = ldap_first_entry(ld, result)) == NULL) {
	report(LOG_ERR, "pwcheck_ldap: ldap_first_entry error: %s",
	       ldap_err2string(ldap_result2error(ld, result, NULL)));
	ldap_msgfree(result);
	ldap_unbind(ld);
	return 0; /* This should not fail unless the last condition failed */
    }

/* Get the DN of the entry */

    if ((dn = ldap_get_dn(ld, entry)) == NULL) {
	report(LOG_ERR, "pwcheck_ldap: ldap_get_dn error: %s",
	       ldap_err2string(ldap_result2error(ld, result, NULL)));
	ldap_msgfree(entry);
	ldap_unbind(ld);
	return 0; /* couldnt find dn */
    }

/* Now bind as the DN with the password supplied earlier...
   Successful bind means the password was correct, otherwise the
   password is invalid. */

#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_SASL)
    cred.bv_val = passwd;
    cred.bv_len = strlen(passwd);
    if ((rc = ldap_sasl_bind_s(ld, dn, saslmech,
			       &cred, NULL, NULL, &servcred)) != LDAP_SUCCESS) {
	report(LOG_ERR, "pwcheck_ldap: ldap_sasl_bind_s 2 error: %s", ldap_err2string(rc));
#else
    if ((rc = ldap_simple_bind_s(ld, dn, passwd)) != LDAP_SUCCESS) {
	report(LOG_ERR, "pwcheck_ldap: ldap_simple_bind_s 2 error: %s", ldap_err2string(rc));
#endif /* TAC_LDAP_VERSION3 && TAC_LDAP_SASL */
	free(dn);
	ldap_msgfree(entry);
	ldap_unbind(ld);
	return 0; /* couldnt bind therefore username or password wrong */
    }

    free(dn);
    ldap_msgfree(entry);
    ldap_unbind(ld);
    return 1;
}
