/*
 * advanced tacacs+ server based on Cisco Systems tac_plus code.
 * For Cisco copyright notice, see LICENSE.CISCO
 *
 * This is a copyrighted commercial software; see LICENSE for details.
 */

#include "tac_plus.h"
#include "regex.h"

static  struct  tm      acl_tm[1];
extern  struct  agroup  *acltable[];
static  struct  nasgrp  *nas_head = (struct nasgrp *) 0;

#define ACL_RET(x) { if (rmsg) *msg = ; return(x); }

/*
 * Perform access check. Returns ACL_DENY or ACL_PERMIT, respectively.
 */
int     acl_check(aname, id, msg)
	char    *aname, **msg;
	struct  identity *id;
{
	struct  agroup          *ap;
	struct  aexp            *ae;
	int                     aval;
	char                    *rmsg = NULL;

	*msg = NULL;
	if (!aname)
		return (ACL_PERMIT);
	if (debug & DEBUG_AUTHEN_FLAG)
		report(LOG_DEBUG, "acl_chk aname = %s username %s",
			aname, id->username);
#ifdef UDB
	if (*aname == '%') {    /* need substitution */
		DB      *db = (DB *) cfg_get_udb(id->username);

		if (db) {
			aname = db_replace(id->username, db, aname);
			if (*aname == '%') {
				report(LOG_ERR, "can't subst group name for field %s, user %s",
					aname, id->username);
				return (ACL_DENY);
			}
		}
	}
#endif
	if (!(ap = (struct agroup *) hash_lookup(acltable, aname)))
		return (ACL_DENY);
	if (debug & DEBUG_AUTHEN_FLAG)
		report(LOG_DEBUG, "found group %s in acltab", aname);

	/* Check regular expressions */
	for (ae = ap->aexps; ae; ae = ae->next) {
		aval = 1;
		if (ae->agroup)
		       aval &= acl_check(ae->agroup, id, &rmsg);
		if (ae->user)
		       aval &= !regexec(ae->user, id->username, 0, NULL, 0);
		if (ae->group)
		       aval &= acl_checkgrp(ae->group, id->username);
		if (ae->from)
			aval &= !regexec(ae->from, id->NAC_address, 0, NULL, 0);
		if (ae->tty)
			aval &= !regexec(ae->tty,  id->NAS_port, 0, NULL, 0);
#ifdef DXS1
		if (ae->dxs1)
			aval &= acl_dxs1chk(ae->dxs1, id->username, id->NAS_port);
#endif
		if (ae->nasg)
			aval &= acl_naschk(ae->nasg, id->NAS_name);
		if (ae->tg)
			aval &= acl_timeok(ae->tg);
		if (aval) {             /* got a match! */
			if (ap->msg || rmsg)
				*msg = rmsg ? rmsg : ap->msg;
			return (ae->op);
		}
	}
	/* no luck - return default action */
	if (ap->msg || rmsg)
		*msg = rmsg ? rmsg : ap->msg;
	return (ap->defacc);
}

/*
 * Get pointer to NAS structure by name
 */
struct nasgrp  *acl_getnas(name)
	char    *name;
{
	struct nasgrp   *np;

	for (np = nas_head; np; np = np->next)
		if (!strcmp(np->name, name))
			return (np);
	return NULL;
}

/*
 * Free entries in agroup
 */
void    acl_free_agroup(ap)
	struct  agroup  *ap;
{
	struct  aexp    *ae;
	struct  tgrade  *tg;

	if (ap->name)
		free(ap->name);
	if (ap->msg)
		free(ap->msg);
	while (ap->aexps) {
		ae = ap->aexps;
		if (ae->user)
			free(ae->user);
		if (ae->group)
			free(ae->group);
		if (ae->agroup)
			free(ae->agroup);
		if (ae->tty)
			free(ae->tty);
#ifdef DXS1
		if (ae->dxs1)
			free(ae->dxs1);
#endif
		if (ae->from)
			free(ae->from);
		while (ae->tg) {
			tg = ae->tg->next;
			free(ae->tg);
			ae->tg = tg;
		}
		ae = ap->aexps->next;
		free(ap->aexps);
		ap->aexps = ae;
	}
}

/*
 * Parse NAS configuration entry
 */
int     acl_parse_nas(name, s)
	char            *name, *s;
{
	struct nasgrp   *np;
	struct nasl     *nl;
	char            *p;

	/* try to find entry -- if nonexistent, create */
	if (!(np = acl_getnas(name))) {
		np = (struct nasgrp *) tac_malloc (sizeof (*np));
		np->name = (char *) tac_strdup(name);
		np->naslist = np->tail = (struct nasl *) 0;
		np->next = nas_head;
		nas_head = np;
	}
	if ((p = strchr(s, '/')))        /* mask present */
		*p++ = '\0';

	nl = (struct nasl *) tac_malloc(sizeof (*nl));
	nl->imask = 0;
	nl->next = (struct nasl *) 0;

	if (!(nl->ipaddr = inet_addr(s)))
		return 1;
	if (p) {
		if (!(nl->imask = inet_addr(p)))
			return 1;
	} else
		nl->imask = 0;
	nl->imask  = ~nl->imask;
	nl->ipaddr &= nl->imask;
	/* got both address & mask; now add to list STARIGHTFORWARD */
	if (!np->tail) {
		np->naslist = np->tail = nl;
		return 0;
	}
	np->tail->next = nl;
	np->tail = nl;
	return 0;
}

/*
 * Check NAS address list
 */
int     acl_naschk(nasg, s)
	struct nasgrp  *nasg;
	char           *s;
{
	struct nasl    *np;

	u_long  a = inet_addr(s);

	if (!a)
		return (ACL_DENY);
	for (np = nasg->naslist; np; np = np->next)
		if ((a & np->imask) == np->ipaddr)
			return (ACL_PERMIT);
	return (ACL_DENY);
}

/*
 * Free NAS lists
 */
void    acl_free_nasgrp()
{
	struct  nasgrp  *ng1, *ng2;
	struct  nasl    *nl1, *nl2;

	ng1 = nas_head;
	while (ng1) {
		if (ng1->name)
			free(ng1->name);
		nl1 = ng1->naslist;
		while (nl1) {
			nl2 = nl1->next;
			(void) free (nl1);
			nl1 = nl2;
		}

		ng2 = ng1->next;
		(void) free(ng1);
		ng1 = ng2;
	}
	nas_head = (struct nasgrp *) NULL;
}

/*
 * Time-grades stuff ----------------------------------------------------
 */
struct  wklist {
	char    *abbrev;                /* Weekday abbreviation         */
	int     alen;                   /* abbreviation length          */
	int     wmin;                   /* Day from                     */
	int     wmax;                   /* Day until                    */
};

static struct wklist wklist[] = {
    { "any", 3, 0, 6 },         { "wk", 2, 1, 5 },      { "su", 2, 0, 0 },
    { "mo",  2, 1, 1 },         { "tu", 2, 2, 2 },      { "we", 2, 3, 3 },
    { "th",  2, 4, 4 },         { "fr", 2, 5, 5 },      { "sa", 2, 6, 6 },
    { "never", 5, -1, -2 },     { "none", 4, -1, -2 },  { NULL, 0, 0, 0 }
};

/* Set and save current time */
void    acl_settime()
{
	time_t          t;

	(void) time(&t);
	*acl_tm = *localtime(&t);
}

/*
 * parse timegrade string
 */
struct  tgrade *acl_parse_tgrade(s)
	char    *s;
{
	struct wklist  *wp;
	struct tgrade  *tlist, *ttail, *tp;
	int             n1, n2, op;
	char           *p;

	ttail = tlist = (struct tgrade *) 0;

	for(;;) {
		if (*s == '!') {
			op = 1;
			s++;
		} else
			op = 0;
		for (wp = wklist; wp->alen; wp++)
			if (!strncmp(s, wp->abbrev, wp->alen))
				break;
		if (!wp->alen)
			return NULL;
		s += wp->alen;
		if (!*s || *s == ',') {
			n1 = 0;         n2 = 2359;
			goto done;
		}
		if (!(p = (char *) strchr(s, '-')))
			return NULL;
		*p = '\0';
		if (sscanf(s, "%d", &n1) != 1)
			return NULL;
		for (; s < p; s++)
			if (*s < '0' || *s > '9')
				return NULL;
		if (!*(s = ++p))
			return NULL;
		while (*s && *s >= '0' && *s <= '9')
			s++;
		if (sscanf(p, "%d", &n2) != 1 || (*s && *s != ','))
			return NULL;
done:
		tp = (struct tgrade *) tac_malloc(sizeof(*tp));
		tp->next = (struct tgrade *) 0;
		tp->wkfrom = wp->wmin;   tp->wkto   = wp->wmax;
		tp->mfrom  = ((n1 / 100) * 60) + (n1 % 100);
		tp->mto    = ((n2 / 100) * 60) + (n2 % 100);
		tp->rev     = op;
		if (tlist) {
			ttail->next = tp;
			ttail = tp;
		} else
			ttail = tlist = tp;
		if ((n1 % 100) > 59 || (n2 % 100 > 59) ||
		    tp->mfrom > 1439 || tp->mto > 1439)
			return NULL;
		if (!*s)
			return (tlist);
		s++;
	}
}

/*
 * Check if time is right
 */
int     acl_timeok(tp)
	struct tgrade  *tp;
{
	int             t, n1, n2, inv;
	int             permit = 1, nrev = 0;

	if (!tp)
		return ACL_PERMIT;
	for (; tp; tp = tp->next) {
		if (tp->rev)
			nrev++;
		if (acl_tm->tm_wday < tp->wkfrom ||
		    acl_tm->tm_wday > tp->wkto) {
			permit &= tp->rev;
			continue;
		}
		if (tp->mfrom <= tp->mto) {
			n1 = tp->mfrom; n2 = tp->mto;   inv = 0;
		} else {
			n1 = tp->mto;   n2 = tp->mfrom; inv = 1;
		}
		t = (acl_tm->tm_hour * 60) + acl_tm->tm_min;
		t  = (t >= n1 && t <= n2);
		n1 = (t ^ inv) ^ tp->rev;
		if (!t) {
			permit &= n1;
			continue;
		}
		return (n1);
	}
	return (nrev ? permit : ACL_DENY);
}

/*
 * Verify auth data against ACL's
 */
int     acl_verify(name, data, msg)
	char    *name, **msg;
	struct  authen_data *data;
{
    char        *s;

    if (debug & DEBUG_AUTHEN_FLAG)
	    report(LOG_DEBUG, "acl_verify: name = <%s>", name);
    s = cfg_get_pvalue(name, TAC_IS_USER, S_agroup, TAC_PLUS_RECURSE);
    acl_settime();
    if (acl_check(s, data->NAS_id, msg) == ACL_DENY) {
	    if (debug & DEBUG_AUTHEN_FLAG)
		report(LOG_DEBUG, "acl_verify DENY name %s", name);
	    return (0);
    }
    if (debug & DEBUG_AUTHEN_FLAG)
	report(LOG_DEBUG, "acl_verify ret PERMIT");
    return (1);
}

#ifdef DXS1
regex_t dxs1_re;
int acl_dxs1chk(regex_t *dxs1, char *username, char *NAS_port) {
  regmatch_t pmatch[2];
  char *buf;
  char cmd[256];
  int res = ACL_PERMIT; /* FIXME default behavior */

  report(LOG_DEBUG, "dxs1 check user=`%s' port=`%s'", username, NAS_port);
  if (!strncmp(NAS_port, "Async", 5) || !strncmp(NAS_port, "tty", 3))
    while (!isdigit(*NAS_port) && *NAS_port) ++NAS_port;
  snprintf(cmd, sizeof(cmd)-1, PATH_RSH " '%s'%s%s 'show line %s'",
      session.peer,
      (session.nas && session.nas->rshuser) ? " -l " : "",
      (session.nas && session.nas->rshuser) ? session.nas->rshuser : "",
      NAS_port);

  report(LOG_DEBUG, "dxs1: run %s", cmd);

  if ((buf = call_rsh(cmd))) {
    if (!regexec(&dxs1_re, buf, 2, pmatch, 0)) {
      *(buf+pmatch[1].rm_eo) = '\0';
      res = !regexec(dxs1, buf+pmatch[1].rm_so, 0, NULL, 0);
      if (!res)
        report(LOG_DEBUG, "acl_dxs1chk: dxs1=`%s' not matched for %s",
	       buf+pmatch[1].rm_so, username);
    } else
      report(LOG_DEBUG, "dxs1 not found user=%s, nas=%s, line=%s",
	     username, session.peer, NAS_port);
    free(buf);
  } else report(LOG_ERR, "acl_dxs1chk: error reading rsh data user=%s, nas=%s, line=%s",
		username, session.peer, NAS_port);
  return (res);
}
#endif
