/*
 * advanced tacacs+ server based on Cisco Systems tac_plus code.
 * For Cisco copyright notice, see LICENSE.CISCO
 *
 * This is a copyrighted commercial software; see LICENSE for details.
 */

#include "tac_plus.h"
#include <sys/stat.h>

#define MAXUDB          16              /* max. num. of opened databases */
#define MAXFLDS         10              /* fields in database            */
#define MAXFLEN         256             /* max. length of string         */

static struct dreg {
	char    *dbname;        /* database name   */
	int     refcnt;         /* reference count */
	DB      *hdl;           /* existing handle */
	struct  stat st;
} dregs[MAXUDB];

/*
 * Register (and open) new database
 */
DB *db_register(dbname)
	char    *dbname;
{
	DB              *hdl;
	struct dreg     *dp;

	for (dp = dregs; dp < &dregs[MAXUDB]; dp++)
		if (dp->hdl && !strcmp(dbname, dp->dbname))
				return dp->hdl;
	/* find empty slot */
	for (dp = dregs; dp < &dregs[MAXUDB]; dp++)
		if (!dp->hdl)
			break;
	if (dp >= &dregs[MAXUDB]) {
		report(LOG_ERR, ": too many different databases (>%d)",
			MAXUDB);
		return NULL;
	}
	hdl = (DB *) dbopen(dbname, O_RDONLY, 0555, DB_HASH, NULL);
	if (!hdl) {
		report(LOG_ERR, ": cannot open database %s", dbname);
		return NULL;
	}
	if (!tac_lockfd(dbname, (*hdl->fd)(hdl), 0)) {
	  report(LOG_ERR, "unable to lock database %s: %s", dbname, strerror(errno));
	  return NULL;
	}
	dp->dbname = tac_strdup(dbname);
	fstat((*hdl->fd)(hdl), &dp->st);
	return (dp->hdl = hdl);
}

char    *db_getname(hdl)
	DB      *hdl;
{
	struct dreg     *dp;

	for (dp = dregs; dp < &dregs[MAXUDB]; dp++)
		if (hdl == dp->hdl)
			return dp->dbname;
	return "UNKNOWN";
}

/*
 * Perform replacements for given user in string 's'; return result
 * in static array (or pointer to original argument, if unchanged)
 */
char *db_replace(user, db, s)
	DB     *db;
	char   *user, *s;
{
	DBT     k[1], d[1];
	char    flds[MAXFLDS][MAXFLEN], *p, *r, *ep;
	static  char os[MAXFLEN];
	typeof  (flds[0])  *fp;
	int     ret;
#if defined(AUTO_MEMBERS) && defined(AM_REGEXP) && defined(MORE_SUBST)
	char db_user[MAX_INPUT_LINE_LEN], *db_user1;
#endif

	if (!db)
		return s;       /* nothing to do - return source */

#if defined(AUTO_MEMBERS) && defined(AM_REGEXP) && defined(MORE_SUBST)
	db_user1 = cfg_get_pvalue(user, TAC_IS_USER, S_db_user, TAC_PLUS_RECURSE);
	if (db_user1) {
	  regex_t *users;
	  char *db_user2;
	  struct identity id;
	  id.username = user;
	  if ((users = cfg_get_pvalue(user, 1, S_users, TAC_PLUS_RECURSE)) &&
	      !regexec(users, user, NMATCH, id.pmatch, 0)) {
	    db_user2 = substitute3(db_user1, (void*)&id, 2);
	    report(LOG_INFO, "changed db_user from `%s' to `%s'", user, db_user2);
	    strncpy(db_user, db_user2, MAX_INPUT_LINE_LEN - 1);
	    free(db_user2);
	    user = db_user;
	  }
	}
#endif

	k->data = user;
	k->size = strlen(k->data);
	if ((ret = (*db->get) (db, k, d, 0)) < 0) {
		report(LOG_ERR, " %d getting key %s from db %s",
			errno, user, db_getname(db));
		return s;
	} else if (ret == 1) {
		report(LOG_DEBUG, ":%s: no such user in db %s", user,
			db_getname(db));
		return s;
	}

	for (fp = flds; fp < &flds[MAXFLDS]; fp++)
		**fp = '\0';
	fp = flds;
	ep = &((char *) d->data)[d->size];

	/* parse database fields. Fields are separated by '|' */
	for (p = (char *) d->data; *p && p < ep; ) {
		while(*p == '|' && p < ep) {
			p++;
			if (fp >= &flds[MAXFLDS])
				report(LOG_ERR, "too many fields in record %s of db %s",
					user, db_getname(db));
			else
				fp++;
		}
		r = *fp;
		while(*p != '|' && *p && p < ep)
			*r++ = *p++;
		*r++ = '\0';
	}

	/* fields are now in flds[] array - now perform substs */

	r = os;

	for (p = s; *p; p++) {
		if (*p == '%') {        /* some substitution ... */
			p++;
			if (*p == '%') {
				*r++ = '%';
				continue;
			}
			if (*p >= '0' && *p <= '9') {
				ep = flds[(int)(*p - '0')];
				while(*r++ = *ep++)
					;
				r--;
				continue;
			}
			*r++ = '%';     *r++ = *p;
			continue;
		}
		*r++ = *p;
	}
	*r++ = '\0';

	if (debug & DEBUG_SUBST_FLAG)
	  report(LOG_DEBUG, "db_replace user=`%s', `%s' -> `%s'", user, s, os);
	return (os);
}

/*
 * Close all database descriptors..
 */
void    db_closeall()
{
	struct dreg     *dp;

	for (dp = dregs; dp < &dregs[MAXUDB]; dp++)
		if (dp->hdl) {
			(*dp->hdl->close) (dp->hdl);
			dp->hdl = (DB *) 0;
			free(dp->dbname);
		}
}

/*
 * Check if file descriptor is used by some db
 */
int     db_inuse(desc)
	int     desc;
{
	struct  dreg    *dp;

	for (dp = dregs; dp < &dregs[MAXUDB]; dp++)
		if (dp->hdl && (*dp->hdl->fd) (dp->hdl) == desc)
			return (1);
	return 0;
}

/*
 * Check for databases with changed mtimes and re-open them
 */
void    db_checkmt()
{
	struct dreg     *dp;
	struct stat     st;

	for (dp = dregs; dp < &dregs[MAXUDB]; dp++) {
		if (!dp->hdl || !dp->dbname)
			continue;
		if (stat(dp->dbname, &st) < 0) {
			report(LOG_ERR, ": cannot stat db %s !", dp->dbname);
			/* use old db :-/ */
			free(dp->dbname);
			dp->dbname = NULL;
			continue;
		}
		if (st.st_ino != dp->st.st_ino ||
		    st.st_mtime != dp->st.st_mtime) {
			dp->st = st;
			(*dp->hdl->close) (dp->hdl);
			dp->hdl = (DB *) dbopen(dp->dbname, O_RDONLY, 0555, DB_HASH, NULL);
			if (!dp->hdl) {
				report(LOG_ERR, ": can't reopen db %s!", dp->dbname);
				free(dp->dbname);
				dp->dbname = NULL;
			}
			if (!tac_lockfd(dp->dbname, (dp->hdl->fd)(dp->hdl), 0)) {
			  report(LOG_ERR, "unable to lock database %s: %s", dp->dbname, strerror(errno));
			  /* return; */
			}
		}
	}
}
