/*
 * advanced tacacs+ server based on Cisco Systems tac_plus code.
 * For Cisco copyright notice, see LICENSE.CISCO
 *
 * This is a copyrighted commercial software; see LICENSE for details.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <sys/types.h>
#include <sys/socket.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#include <arpa/nameser.h>
#include <setjmp.h>
#include <netdb.h>
#include <errno.h>
#include <stdlib.h>
#include <stdarg.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>
#ifdef HAVE_DB_185_H
#include <db_185.h>
#else
#include <db.h>
#endif
#include "dynipd.h"
#include "allow.h"

extern int sockread(int, u_char *, int, int);

#define NHASH           128
#define NAPOOLS         64
#define MAXACCTTIME     20

struct htab {                           /* two-way hashtable      */
	struct  ud      *udp;           /* pointer to user struct */
	struct  htab    *next;          /* chain link             */
	int             active;         /* active flag            */
}
	*htabi[NHASH],                  /* references by ipns     */
	*htabn[NHASH];                  /* references by names    */

#define H_GET           0001            /* lookup                 */
#define H_UPDATE        0002            /* create/update          */
#define H_DELETE        0004            /* delete entry           */

/* Common structure for in-core user data */
struct  ud {
	time_t          time;           /* allocation time    */
	u_long          addr;           /* assigned address   */
	char           *name;           /* User name          */
	char           *pname;          /* pool name ref      */
	char            nasinfo[32];    /* nas ip/port */
	int		stat;		/* status (0: free, 1: given, 2: in use) */
	time_t          timestamp;      /* time when stat=1 */
};

struct dqueue {                         /* Database update stack          */
	struct  ud     *udp;            /* pointer to user structure      */
	int             op;             /* Operation                      */
	struct dqueue  *next;
	struct dqueue  *prev;
};

static struct dqueue *dq_head = (struct dqueue *) 0,
		     *dq_tail = (struct dqueue *) 0;

struct apool {
	u_long          a1;             /* Beginning of range */
	u_long          a2;             /* End of range       */
	int             ps;             /* Range size         */
	u_long         *stack;          /* freelist stack     */
	int             sp;             /* stack ptr          */
	struct plist   *pool;           /* pool backref       */
};

struct plist {                          /* Named pools list    */
	int             curpool;        /* for circular search */
	char           *pname;          /* pool list name      */
	int             nranges;        /* number of pools     */
	struct apool    apools[NAPOOLS];/* apool list          */
	struct plist   *next;           /* ptr to next element */
};
static struct plist *p_list = (struct plist *) 0;

/* some defaults */
#ifndef CONFPATH
#define CONFPATH "/etc/tac_plus"
#endif
#ifndef DATAPATH
#define DATAPATH "/var/adm/tac_plus"
#endif
#ifndef LOGPATH
#define LOGPATH DATAPATH
#endif

#ifndef CDBNAME
#define CDBNAME DATAPATH "/ipaddr.db"
#endif
#ifndef CCONFNAME
#define CCONFNAME CONFPATH "/dynipd.conf"
#endif
#ifndef DYNLOGFILE
#define DYNLOGFILE LOGPATH "/dynipd.log"
#endif

#define PTMAX           5               /* secs - protocol timeout */

#define NTOA(s,n)       sprintf((s), "%ld.%ld.%ld.%ld",  \
	((n) >> 24) & 0xFF, ((n) >> 16) & 0xFF,                 \
	((n) >> 8 ) & 0xFF, (n) & 0xFF)

#define SANE_RANGE      16384           /* how many addresses ...    */

static DB       *cachedb = NULL;        /* cache DB handle         */
static char     cachedbname[512] = "";  /* cache DB name           */
static int      debug   = 0;            /* debugging               */
static int      inter   = 0;            /* non-background          */
static int      udb_update = 15;        /* dbupdate - 30 secs dflt */
static char     confname[FILENAME_MAX] = CCONFNAME;
static u_long   Seq;                    /* Sequence #              */
static int      dynipd_port = 0;        /* daemon port number      */
static int      proto_clock = -1,
		upd_clock   = 0;
static int      my_pid;
sigjmp_buf      abrt;
static char     Logfile[512] = DYNLOGFILE;
static char     default_pool_name[] = "default";

void    parse_cf(void), report(int, char *, ...), doexit(int);
int     dyna_setrange(char *, char *);
struct  ud *hashi(int, u_long, struct ud *),
	*hashn(int, char *, struct ud *, char *);
void    *lmalloc(int);
char    *lstrdup(const char *);
u_long  dyna_alloc(char *, struct plist *, char *), dyna_lalloc(struct plist *);
void    udb_sync(), udb_load();
#ifdef HARSH_POLICY
void dyna_invalidate(struct plist *);
#endif
RETSIGTYPE myclock(int);
void    udb_queue_request(int, struct ud *);
struct  apool *getap(u_long);

#if 0 /* FIXME only for debug */
RETSIGTYPE dumphash(int s) {
  struct  htab     *hp;
  u_long  l;
  struct in_addr in;

  report(1, "--- hash tables ---");
  report(1, "nhash:");
  for (l = 0; l < NHASH; l++) {
    for (hp = htabn[l]; hp; hp = hp->next) {
      in.s_addr = hp->udp->addr;
      report(1, "-%d %d<%s> %s %d (%s)-", hp->active, hp->udp->stat, hp->udp->name,
	     inet_ntoa(in), hp->udp->time, hp->udp->nasinfo);
    }
  }
  report(1, "ihash:");
  for (l = 0; l < NHASH; l++) {
    for (hp = htabi[l]; hp; hp = hp->next) {
      in.s_addr = hp->udp->addr;
      report(1, "-%d %d<%s> %s %d (%s)-", hp->active, hp->udp->stat, hp->udp->name,
	     inet_ntoa(in), hp->udp->time, hp->udp->nasinfo);
    }
  }
  report(1, "------------------------------");
  signal(SIGUSR2, dumphash);
}
#endif

/*
 * Options:
 *      -g              - debug mode; don't go to background
 *      -I              - interactive mode; don't go to background
 *      -c <confname>   - alternate location of config. file
 *      -D <dbname>     - altername cachedb location, overrides cf
 *      -L <logfile>    - alternate log location
 *      -p port         - listen on alternate port
 */

int main(argc, argv)
	int     argc;
	char    *argv[];
{
	int                     i, sc, sd, one = 1;
	u_long  addr;
	size_t	psalen;
	struct  sockaddr_in     isa, psa;
	struct  itimerval       it[1];
	struct  apool           *ap;
	struct  plist           *pp;
	struct dynp             dpb;
	char                    username[MAXARGBUF/2],
				poolname[MAXARGBUF/2],
				nasinfo[32];

	while((i = getopt(argc, argv, "gIc:D:p:L:")) != -1) {
		switch (i) {
			case 'g':
				debug = 1;
				continue;
			case 'I':
				inter = 1;
				continue;
			case 'c':
				strcpy(confname, optarg);
				continue;
			case 'D':
				strcpy(cachedbname, optarg);
				continue;
			case 'L':
				strcpy(Logfile, optarg);
				continue;
			case 'p':
				dynipd_port = atoi(optarg);
				continue;
			case '?':
usage:
				report(2, "Usage: %s [-g] [-I] [-c conf] [-D cachedb]",
					argv[0]);
		}
	}
	my_pid = getpid();
	if (optind < argc)
		goto    usage;

	parse_cf();                     /* read config */

	/* Initialize stacks */
	for (pp = p_list; pp; pp = pp->next) {
		for (ap = pp->apools; ap < &pp->apools[pp->nranges]; ap++) {
			ap->stack = (u_long *)
			    lmalloc((ap->ps + 1) * sizeof (u_long));
			ap->sp    = -1;
		}
	}
	/* Daemonize */
	if (!debug && !inter) {
		if (fork())
			exit(0);
		setsid();
	}

	if (!dynipd_port)
		dynipd_port = DPORT;

	signal(SIGTERM, doexit);
	signal(SIGINT,  doexit);
	signal(SIGALRM, myclock);
	signal(SIGCHLD, SIG_IGN);
#if 0 /* FIXME only for debug */
	signal(SIGUSR2, dumphash);
#endif

	/* set timer */
	it->it_value.tv_sec  = 1;
	it->it_value.tv_usec = 1;
	it->it_interval = it->it_value;

	setitimer(ITIMER_REAL, it, NULL);         /* launch clock */

	udb_load();                     /* load current map */

#ifdef DEBUG
	testloop();
#endif
	/* Do all socket stuff */
	if ((sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
		report(2, "can't get socket");

	if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR,
	    (char *) &one, sizeof (one)) < 0)
		report(2, "setsockopt REUSEADDR: %s", strerror(errno));

	if (setsockopt(sd, SOL_SOCKET, SO_KEEPALIVE,
	    (char *) &one, sizeof (one)) < 0)
		report(2, "setsockopt KEEPALIVE: %s", strerror(errno));

	bzero((char *) &isa, sizeof (isa));
	isa.sin_family = AF_INET;
	isa.sin_addr.s_addr = htonl(INADDR_ANY);
	isa.sin_port = htons(dynipd_port);

	if (bind(sd, (struct sockaddr *) &isa, sizeof (isa)) < 0)
		report(2, "can't bind to port %d: %s",
			dynipd_port, strerror(errno));

	if (listen(sd, 10) < 0)
		report(2, "listen returned error: %s",strerror(errno));

	report(1, "Daemon mode; listening on port %d", dynipd_port);

	sc = -1;
	for (;;) {
		sigsetjmp(abrt, 1);
		if (sc > 0)
			close(sc);
		psalen = sizeof(psa);
		if ((sc = accept(sd, (struct sockaddr *) &psa, &psalen)) < 0) {
#ifndef BSD
			if (errno==EINTR) continue;
#endif
			report(1, "accept() failed, %s", strerror(errno));
			continue;
		}
		if(allowaddr(psa.sin_addr.s_addr))
			report(0, "accepted call from %s",
				inet_ntoa(psa.sin_addr));
		else {
			report(1, "Unauthorized call from %s",
				inet_ntoa(psa.sin_addr));
			continue;
		}
		proto_clock = 0;
		read(sc, (char *) &dpb, 1);
		if (sockread(sc, (u_char *) &dpb, sizeof (dpb), 5) == -1) {
			report(1, "protocol error: %s", strerror(errno));
			continue;
		}

		if (dpb.magic != DYNP_MAGIC ||
		    (dpb.op != DYNP_IARQ && dpb.op != DYNP_IARQ2
		     && dpb.op != DYNP_LONOT && dpb.op != DYNP_LINOT)) {
			report(1, "bad magic in incoming packet");
			continue;
		}
		write(sc, &one, 1);      /* ACK */
		proto_clock = -1;

		strncpy(username, dpb.argb, (MAXARGBUF/2) - 1);
		switch (dpb.op) {
		case DYNP_IARQ:
		  strcpy(poolname, default_pool_name);
		  strncpy(nasinfo, &dpb.argb[MAXARGBUF/2], 32);
		  break;
		case DYNP_IARQ2:
		  strncpy(poolname, &dpb.argb[MAXARGBUF/2], (MAXARGBUF/2) - 1);
		  strncpy(nasinfo, &dpb.argb[MAXARGBUF/2] + strlen(poolname) + 1, 32);
		  break;
		case DYNP_LONOT:
		  addr = htonl(inet_addr(&dpb.argb[MAXARGBUF/2]));
		  if (addr == -1) {
		    sprintf(dpb.argb, "LONOT: malformed address");
		    report(1, dpb.argb);
		    dpb.op = DYNP_IERROR;
		  } else {
		    struct ud *udp;
		    if ((udp = hashi(H_GET, addr, NULL)) && udp->stat) {
		      udp->stat = 0;
		      udp->nasinfo[0]=0;
		      report(1, "LONOT: user=%s, ip=%s", username, &dpb.argb[MAXARGBUF/2]);
		      dpb.op = DYNP_OKAY;
		    } else {
		      if (getap(addr)) {
			sprintf(dpb.argb, "LONOT: user=%s, address %s is unused", username, &dpb.argb[MAXARGBUF/2]);
			report(1, dpb.argb);
			dpb.op = DYNP_IERROR;
		      } else dpb.op = DYNP_OKAY;
		    }
		  }
		  goto reply;
		case DYNP_LINOT:
		  addr = htonl(inet_addr(&dpb.argb[MAXARGBUF/2]));
		  if (addr == -1) {
		    sprintf(dpb.argb, "LINOT: malformed address");
		    report(1, dpb.argb);
		    dpb.op = DYNP_IERROR;
		  } else {
		    struct ud *udp;
		    strncpy(nasinfo, &dpb.argb[MAXARGBUF/2] + strlen(&dpb.argb[MAXARGBUF/2]) + 1, 32);
		    dpb.op = DYNP_OKAY;
		    if ((udp = hashi(H_GET, addr, NULL))) {
		      udp->stat = 2;
		      strncpy(udp->nasinfo, nasinfo, 32); /* not needed really */
		      /* FIXME check if udp->name is the same as username;
		       * otherwise remove old user and add a new one
		       */
		      report(1, "LINOT: user=%s, ip=%s", username, &dpb.argb[MAXARGBUF/2]);
		    } else {
		      struct apool *ap;
		      if ((ap = getap(addr))) {
			struct ud uda[1];
			uda->addr  = addr;
			uda->name = lstrdup(username);
			uda->time  = Seq++;
			uda->pname = ap->pool->pname;
			strncpy(uda->nasinfo, nasinfo, 32);
			uda->timestamp = time(NULL);
			uda->stat = 2;

			udb_queue_request(H_UPDATE, uda);
			hashn(H_UPDATE, username, uda, NULL);

			sprintf(dpb.argb, "LINOT: add user=%s, ip=%s", username, &dpb.argb[MAXARGBUF/2]);
			report(1, dpb.argb);
		      }
		    }
		  }
		  goto reply;
		}
		report(1, "IARQ: %s [%s] (%s)", username, poolname, nasinfo);
		for (pp = p_list; pp; pp = pp->next)
			if (!strcmp(pp->pname, poolname))
				break;
		if (pp) {
			u_long l;
			switch(l = dyna_alloc(username, pp, nasinfo)) {
			case 0:
			    sprintf(dpb.argb, "no db entry found for %s [static]", username);
			    dpb.op = DYNP_IERROR;
			    report(1, dpb.argb);
			    break;
#ifndef HARSH_POLICY
			case 1:
			    sprintf(dpb.argb, "unable to allocate address for %s [%s]", username, pp->pname);
			    dpb.op = DYNP_IERROR;
			    report(1, dpb.argb);
			    break;
#endif
			default:
			    NTOA(dpb.argb, l);
			    dpb.op = DYNP_IAREPL;
			    report(1, "alloc: %s <- %s [%s]", dpb.argb, username, poolname);
			}
		} else {
			sprintf(dpb.argb, "Pool %s not found", poolname);
			dpb.op = DYNP_IERROR;
			report(1, dpb.argb);
		}
reply:		proto_clock = 0;
		if (write(sc, (char *) &dpb, sizeof (dpb)) != sizeof (dpb)) {
			report(1, "error replying to request");
			continue;
			/* we should better deallocate address here ... */
		}
		proto_clock = -1;
	}
}

/*
 * Allocate address for particular user
 */
u_long  dyna_alloc(user, pool, nasinfo)
char           *user;
struct plist   *pool;
char           *nasinfo;
{
	struct ud       *udp, uda[1];
	struct apool    *ap;

	/* check in hashtab first */
	while((udp = hashn(H_GET, user, NULL, nasinfo))) {
		if (udp->pname == pool->pname) {
			udp->time = Seq++;                  /* refresh */
			udp->stat = 1;
			strncpy(udp->nasinfo, nasinfo, 32);
			udp->timestamp = time(NULL);
			udb_queue_request(H_UPDATE, udp);
			return (udp->addr);
		}
		if (!(ap = getap(udp->addr)))
			report(2, "internal err: addr %ld not in any range",
			    udp->addr);
		ap->stack[++ap->sp] = udp->addr;
		udb_queue_request(H_DELETE, udp);
		hashn(H_DELETE, user, udp, NULL);
		report(1, "info: pool changed %s -> %s for user %s", udp->pname, pool->pname, user);
	}
	if (!strncmp(pool->pname, "static", 6)) return 0;
	/* not in cache -- allocate locally and create new entry */
	uda->addr  = dyna_lalloc(pool);
#ifndef HARSH_POLICY
	if (!(uda->addr)) return 1;
#endif
	uda->name = lstrdup(user);
	uda->time  = Seq++;
	uda->pname = pool->pname;
	strncpy(uda->nasinfo, nasinfo, 32);
	uda->timestamp = time(NULL);
	uda->stat = 1;

	udb_queue_request(H_UPDATE, uda);
	hashn(H_UPDATE, user, uda, NULL);

	return (uda->addr);
}

/*
 * Get next number from local stack
 */
u_long  dyna_lalloc(pool)
struct plist    *pool;
{
	struct  apool   *ap;
	int     oldc = pool->curpool;

	for (;;) {
		for (ap = &pool->apools[pool->curpool];
		    pool->curpool < pool->nranges; ap++, pool->curpool++)
			if (ap->sp >= 0)
				return (ap->stack[ap->sp--]);
		pool->curpool = 0;
		for (ap = pool->apools;
		    pool->curpool < oldc; ap++, pool->curpool++)
			if (ap->sp >= 0)
				return (ap->stack[ap->sp--]);
#ifdef HARSH_POLICY
		dyna_invalidate(pool); /* _should_ add at least one element */
#else
		return 0;
#endif
	}
}

/*
 * Read and parse configuration file.
 * Configuration file syntax:
 *
 * db     <dbfilename>            - define db for cache
 * range  x.x.x.x y.y.y.y         - define pool range; multiple allowed
 * update <seconds>               - db update interval in secs
 * allow-from <ipaddr>            - accept connections only from given host
 */
void    parse_cf()
{
	FILE    *cfp;
	char    *p, *q, iline[128];
	int     lnum = 0;

	if (!(cfp = fopen(confname, "r")))
		report(2, "can't open config file %s", confname);
	while(fgets(iline, 128, cfp)) {
		lnum++;
		for (p = iline; *p != '\n' && *p != '#' && *p; p++)
			;
		*p = '\0';
		if (!*iline)
			continue;
		for (p = iline; *p != ' ' && *p != '\t' && *p; p++)
			;
		if (!*p)
synt:
			report(2, "syntax error in config file line %d", lnum);
		*p = '\0';
		for (q = ++p; *q == ' ' || *q == '\t'; q++);
			;
		if (!*q)
			goto synt;
		if (!strcmp(iline, "db")) {
			strcpy(cachedbname, q);
		} else if (!strcmp(iline, "update")) {
			if (sscanf(q, "%d", &udb_update) != 1)
				goto synt;
			if (udb_update < 5)
				report(2, "too small update period");
		} else if (!strcmp(iline, "range")) {
			if (dyna_setrange(default_pool_name, q))
				goto synt;
		} else if (!strcmp(iline, "pool")) {
			for (p = q; *p != ' ' && *p != '\t' && *p; p++)
				;
			if (!p[0] || !p[1])
				goto synt;
			*p++ = '\0';
			while(*p == ' ' || *p == '\t')
				p++;
			if (dyna_setrange(q, p))
				goto synt;
		} else if (!strcmp(iline, "logfile")) {
			if (!strcmp(Logfile, DYNLOGFILE))
				strcpy(Logfile, q);
		} else if (!strcmp(iline, "allow-from")) {
			if (put_new_allow_addr(q)) {
				report(1, "bad ip address %s", q);
				goto synt;
			}
		} else if (!strcmp(iline, "port")) {
			int     port = atoi(q);

			if (port <= 0 || port > 65535) {
				report(1, "bad port %d", port);
				goto synt;
			}
			if (!dynipd_port)
				dynipd_port = port;
		}
	}
	fclose(cfp);
}

/*
 * Load external db into local hash
 */
void    udb_load()
{
	struct  apool   *ap;
	struct  plist   *pp;
	DBT             k[1], d[1];
	struct  ud      udp[1];
	u_long          l;
	int             lfree = 0;

	cachedb = dbopen(cachedbname, O_RDWR, 0600, DB_HASH, NULL);
	if (!cachedb) {
		unlink(cachedbname);
		cachedb = dbopen(cachedbname, O_CREAT|O_RDWR, 0600, DB_HASH, NULL);
		if (!cachedb)
			report(2, "cannot open db %s", cachedbname);
	}

	while(!(*cachedb->seq) (cachedb, k, d, R_NEXT)) {
		udp->name = (char *) lmalloc(k->size + 1 - sizeof(u_long));
		strncpy(udp->name, k->data + sizeof(u_long), k->size - sizeof(u_long));
		udp->name[k->size - sizeof(u_long)]='\0';
		udp->addr = *((u_long *) k->data);
		udp->time = *((time_t *) d->data);
		udp->nasinfo[0] = 0;
		udp->stat = 0;

		if (udp->time > Seq)
			Seq = udp->time;

		if (!(ap = getap(udp->addr))) {
			report(1, "address not from interval: %08X", udp->addr);
			free(udp->name);
			(*cachedb->del) (cachedb, k, NULL);
			continue;
		}
		/* maybe we need to be more discrete here? */
		if (hashi(H_GET, udp->addr, NULL)) {
			report(1, "duplicate addr in db: %08X", udp->addr);
			free(udp->name);
			(*cachedb->del) (cachedb, k, NULL);
		} else {
			udp->pname = ap->pool->pname; /* pool reference */
			hashn(H_UPDATE, udp->name, udp, NULL);
		}
	}
	(*cachedb->sync) (cachedb, NULL);
	Seq++;

	/* Fill the stacks */
	for (pp = p_list; pp; pp = pp->next) {
		lfree = 0;
		for (ap = pp->apools; ap < &pp->apools[pp->nranges]; ap++) {
			ap->sp = -1;
			for (l = ap->a1; l <= ap->a2; l++)
				if (!hashi(H_GET, l, NULL)) {
					ap->stack[++ap->sp] = l;
					lfree++;
				}
		}
#ifdef HARSH_POLICY
		if (lfree <= 0)
			dyna_invalidate(pp);              /* uh oh */
#endif
	}
}

#ifdef HARSH_POLICY
/*
 * compare elements of udp pools
 */
int     dyna_comp(u1, u2)
	const void *u1, *u2;
{
	if (((struct ud*)(u1))->time > ((struct ud*)(u2))->time)
		return (1);
	if (((struct ud*)(u1))->time < ((struct ud*)(u2))->time)
		return (-1);
	return 0;
}

/*
 * Most complex part - cache invalidation algorithm.
 *      - create array of pointers to udb entries
 *      - sort this array by time
 *      - take some elements from the top of array and add it to pool stacks
 */
void    dyna_invalidate(pool)
struct plist    *pool;
{
	struct apool    *ap;
	struct ud       **upool, *udp;
	int             i, top;

	for (ap = pool->apools; ap < &pool->apools[pool->nranges]; ap++) {
		upool = (struct ud **)
			lmalloc(sizeof(struct ud *) * (ap->ps + 1));

		/* fill the pool */
		for (i = 0; i < ap->ps; i++)
			if (!(upool[i] = hashi(H_GET, ap->a1 + i, NULL)))
				report(2, "internal err - dynainv %d/%08X",
					i, ap->a1);
		qsort(upool, ap->ps, sizeof (struct ud *), dyna_comp);
		top = ap->ps / 8;
		if (top < 1)
			top = 1;
		ap->sp = -1;
		for (i = 0; i < top; i++) {
			udp = upool[i];
			ap->stack[++ap->sp] = udp->addr;
			udb_queue_request(H_DELETE, udp);
			hashn(H_DELETE, udp->name, udp, NULL);
		}
	}
	pool->curpool = 0;
}
#endif

/*
 * Hash-table reference by name
 */
struct  ud      *hashn(op, name, udp, nasinfo)
	int             op;
	char           *name;
	struct ud      *udp;
	char           *nasinfo;
{
	int             k;
	unsigned int    hval = 0;
	struct htab    *hp, *hp1 = 0;

	k = strlen(name);
	while (--k >= 0)
		hval += name[k] * (k + 1);
	hval %= NHASH;

	switch (op) {
	case H_GET:
	  for (hp = htabn[hval]; hp; hp = hp->next)
	    if (hp->active && !strcmp(name, hp->udp->name)) {
	      if (hp->udp->stat == 0)
		hp1 = hp; /* addr not used */
	      else if (!strncmp(hp->udp->nasinfo, nasinfo, 32)) /* stat = 1 or 2 */
		break;
	      else if ((hp->udp->stat == 1) &&
		       (time(NULL) - hp->udp->timestamp > MAXACCTTIME) &&
		       !hp1)
		hp1 = hp; /* addr not confirmed */
	    }
	  if (!hp && (hp = hp1)) {
	    if (hp->udp->stat == 1) {
	      struct in_addr in;
	      in.s_addr = hp->udp->addr;
	      report(1, "reusing not confirmed addr %s for %s (delay %i secs)",
		     inet_ntoa(in), hp->udp->name, time(NULL) - hp->udp->timestamp);
	    }
	  }
	  return (hp ? hp->udp : (struct ud *) 0);

	case H_DELETE:
	  for (hp = htabn[hval]; hp; hp = hp->next)
	    if (hp->active && hp->udp == udp) {
	      hp->udp->nasinfo[0]=0;
	      hp->active = 0;
	      hashi(H_DELETE, hp->udp->addr, NULL);
	      break;
	    }
	  return ((struct ud *) 0);

	case H_UPDATE:
	  for (hp = htabn[hval]; hp; hp = hp->next)
	    if (hp->active && !strcmp(name, hp->udp->name) && hp->udp->addr == udp->addr)
	      break;
	  if (hp) {
	    hp->active = 1;
	    hashi(H_DELETE, hp->udp->addr, NULL);
	    hp->udp->time = udp->time;
	    hp->udp->addr = udp->addr;
	    hp->udp->stat = udp->stat;
	    hashi(H_UPDATE, udp->addr, hp->udp);
	    return (hp->udp);
	  }
	  hp = (struct htab *) lmalloc(sizeof (struct htab));
	  hp->udp       = (struct ud *) lmalloc(sizeof (struct ud));
	  hp->udp->name = udp->name;
	  hp->udp->addr = udp->addr;
	  hp->udp->time = udp->time;
	  hp->udp->pname = udp->pname;
	  strncpy(hp->udp->nasinfo, udp->nasinfo, 32);
	  hp->udp->stat = udp->stat;
	  hp->next      = htabn[hval];
	  hp->active    = 1;
	  htabn[hval]   = hp;
	  hashi(H_UPDATE, udp->addr, hp->udp);
	  return (hp->udp);
	}
	return ((struct ud *) 0);
}

/*
 * Hash-table reference by IP number
 */
struct  ud      *hashi(op, inum, udp)
	int             op;
	u_long          inum;
	struct ud      *udp;
{
	struct htab    *hp;
	unsigned int    hval = ((inum * 1333 + 5) * 11) % NHASH;

	for (hp = htabi[hval]; hp; hp = hp->next)
		if (inum == hp->udp->addr)
			break;
	switch (op) {

	    case H_GET:
		return (hp && hp->active ? hp->udp : (struct ud *)0);

	    case H_DELETE:
		if (hp)
			hp->active = 0;
		return ((struct ud *) 0);

	    case H_UPDATE:
		if (hp) {
			hp->active = 1;
			hp->udp = udp;
			return (udp);
		}
		hp = (struct htab *) lmalloc(sizeof (struct htab));
		hp->udp    = udp;
		hp->next   = htabi[hval];
		hp->active = 1;
		htabi[hval] = hp;

		return (udp);
	}
	return ((struct ud *) 0);
}

/*
 * get a pointer to corresponding address range
 */
struct apool *getap(addr)
	u_long          addr;
{
	struct plist    *pp;
	struct apool    *ap;

	for (pp = p_list; pp; pp = pp->next) {
		for (ap = pp->apools; ap < &pp->apools[pp->nranges]; ap++) {
			if (addr >= ap->a1 && addr <= ap->a2)
				return ap;
		}
	}
	return (struct apool *) 0;
}

#define VALID_IPN(n)    ((n) >= 0 && (n) <= 255)

int     dyna_setrange(pname, s)
	char    *pname, *s;
{
	struct  apool   *ap;
	struct  plist   *pp;
	int             i = 0, ib[8];

	/* check for pool "pname" */
	for (pp = p_list; pp; pp = pp->next)
		if (!strcmp(pp->pname, pname))
			break;
	if (!pp) {
		pp = (struct plist *) lmalloc(sizeof(struct plist));
		pp->pname = lstrdup(pname);
		pp->curpool = 0;
		pp->nranges = 0;
		pp->next = p_list;
		p_list = pp;
	}
	ap = &pp->apools[pp->nranges];
	ap->pool = pp;
	if (sscanf(s, "%d.%d.%d.%d %d.%d.%d.%d", &ib[3], &ib[2],
	    &ib[1], &ib[0], &ib[7], &ib[6], &ib[5], &ib[4]) != 8)
		goto synt1;

	for (i = 0; i < 4; i++) {
		if (!VALID_IPN(ib[i]) || !VALID_IPN(ib[i + 4])) {
synt1:
			report(1, "bad ip number(s) in cf file");
			return (1);
		}
		ap->a1 |= (ib[i]     << (i << 3));
		ap->a2 |= (ib[i + 4] << (i << 3));
	}
	ap->ps = ap->a2 - ap->a1;
	if (ap->ps < 0 || ap->ps > SANE_RANGE) {
		report(1, "bad IPN range in cf file");
		return (1);
	}
	ap->ps++;
	if (getap(ap->a1) || getap(ap->a2))
		report(2, "error: address ranges overlap!");
	if (++pp->nranges >= NAPOOLS)
		report(2, "error: too many ranges defined for pool %s",
		    pp->pname);
	return (0);
}

/*
 * database update stuff
 */
void    udb_queue_request(op, udp)
	int             op;
	struct ud       *udp;
{
	struct  dqueue  *dq;

	/* add request to dq_list chain link */
	dq = dq_head;
	dq_head = (struct dqueue *) lmalloc(sizeof(struct dqueue));
	if (!dq)
		dq_tail = dq_head;
	dq_head->next = dq;
	dq_head->prev = (struct dqueue *) 0;
	if (dq)
		dq->prev = dq_head;

	dq_head->udp  = (struct ud *) lmalloc(sizeof (struct ud));
	*(dq_head->udp) = *udp;
	dq_head->udp->name = lstrdup(udp->name);
	dq_head->op   = op;
}

/*
 * Actually update DB
 */
void    udb_doupdate()
{
	struct  dqueue   *dq, *dp;
	time_t  udd[1];
	DBT     k[1], d[1];
	char    tkey[128];

	report(0, "udb_doupdate called");

	if (!dq_tail)
		return;

	for (dq = dq_tail; dq; ) {
		k->size = strlen(dq->udp->name) + sizeof(u_long);
		k->data = tkey;
		* (u_long *) tkey = dq->udp->addr;
		strcpy(k->data + sizeof(u_long), dq->udp->name);

		if (dq->op == H_DELETE) {
			(*cachedb->del) (cachedb, k, NULL);
		} else if (dq->op == H_UPDATE) {
			*udd = dq->udp->time;
			d->data   = udd;
			d->size   = sizeof (time_t);
			if ((*cachedb->put) (cachedb, k, d, NULL) == -1)
				report(1, "can't update db!");
		}
		dp = dq;
		dq = dq->prev;
		free(dp->udp->name);
		free(dp->udp);
		free(dp);
	}
	dq_head = dq_tail = (struct dqueue *) 0;
	(*cachedb->sync) (cachedb, NULL);
}

/*
 * Local time-related events handler
 */
RETSIGTYPE myclock(int disp) {
	signal(SIGALRM, SIG_IGN);

	if (upd_clock++ >= udb_update) {
		upd_clock = 0;
		udb_doupdate();
	}
	if (proto_clock >= 0 && proto_clock++ >= PTMAX) {
		proto_clock = -1;
		report(1, "Session aborted - timeout!");
		signal(SIGALRM, myclock);
		siglongjmp(abrt, 0);
	}
	signal(SIGALRM, myclock);
}

/*
 * Logging stuff
 */
void    report(int lvl, char *fmt,...)
{
	va_list         ap;
	char            msg[1024];      /* longer than any other buffers */
	int             logfd;

	va_start(ap, fmt);
	vsprintf(msg, fmt, ap);
	va_end(ap);

	if (debug) {
	  fputs(msg, stderr);
	  fputs("\n", stderr);
	}

	if (debug || lvl > 0) {
	    logfd = open(Logfile, O_CREAT | O_WRONLY | O_APPEND, 0644);
	    if (logfd >= 0) {
		char buf[1024];
		time_t t;
		int i;

		for(i = 0; i < 10; i++) {
		  if (i > 0) usleep(200000);
		  if (lockf(logfd, F_TLOCK, 0) == 0)
		    break;
		}
		/* if (i < 10) { */
		time(&t);
		strcpy(buf, (char *) ctime(&t));
		buf[24] = ' ';
		sprintf(&buf[25], "[%d]: %s\n", my_pid, msg);
		write(logfd, buf, strlen(buf));
		/* } */
		close(logfd);
	    }
	}
	if (lvl > 1)
		doexit(lvl);
}

/* memory allocator */
void    *lmalloc(size)
	int     size;
{
	void    *p = (void *) malloc(size);

	if (!p)
		report(2, "can't allocate memory, exiting");
	return (p);
}

char *lstrdup(const char *p) {
  char *n = strdup(p);
  if (n == NULL)
    report(2, "strdup allocation failure");
  return (n);
}

/*
 * doexit -- clean exit
 */
void    doexit(n)
	int     n;
{
	if (cachedb) {
		udb_doupdate();
		(*cachedb->close) (cachedb);
	}
	exit(n);
}

#ifdef DEBUG

testloop()
{
	char    *p, *q, ibuf[128], tmp[128];
	struct  htab     *hp;
	struct  plist    *pp;
	u_long  l;

	for (;;) {
#if 0
		printf(" --- hash tables ---\nnhash:");
		for (l = 0; l < NHASH; l++) {
			for (hp = htabn[l]; hp; hp = hp->next)
				printf("-%d<%s> %08X %d-", hp->active, hp->udp->name,
					hp->udp->addr, hp->udp->time);
		}
		printf("\nihash:\n");
		for (l = 0; l < NHASH; l++) {
			for (hp = htabi[l]; hp; hp = hp->next)
				printf("-%d<%s> %08X %d-", hp->active, hp->udp->name,
					hp->udp->addr, hp->udp->time);
		}
		printf("\n------------------------------\n\n");
#endif

		fputs("dynipd> ", stdout);

		if (!fgets(ibuf, 128, stdin))
			doexit(1);
		if (p = strchr(ibuf, '\n')
			*p = 0;
		if (!ibuf[0])
			continue;
		for (q = ibuf; *q != ' ' && *q; q++)
			;
		*q++ = '\0';
		while (*q == ' ')
			q++;
		for (pp = p_list; pp; pp = pp->next)
			if (!strcmp(pp->pname, ibuf))
				break;
		if (!pp) {
			printf("unknown pool %s\n", ibuf);
			continue;
		}
		l = dyna_alloc(q, pp);   NTOA(tmp, l);
		printf("alloc for %s: %s\n", ibuf, tmp);
	}
}
#endif
