/* 
   Copyright (c) 1995-2000 by Cisco systems, Inc.

   Permission to use, copy, modify, and distribute modified and
   unmodified copies of this software for any purpose and without fee is
   hereby granted, provided that (a) this copyright and permission notice
   appear on all copies of the software and supporting documentation, (b)
   the name of Cisco Systems, Inc. not be used in advertising or
   publicity pertaining to distribution of the program without specific
   prior permission, and (c) notice be given in supporting documentation
   that use, modification, copying and distribution is by permission of
   Cisco Systems, Inc.

   Cisco Systems, Inc. makes no representations about the suitability
   of this software for any purpose.  THIS SOFTWARE IS PROVIDED ``AS
   IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
   WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
   FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "tac_plus.h"
#include "expire.h"

/* internal state variables */
#define STATE_AUTHEN_START   0	/* no requests issued */
#define STATE_AUTHEN_GETUSER 1	/* username has been requested */
#define STATE_AUTHEN_GETPASS 2	/* password has been requested */

struct private_data {
    char password[MAX_PASSWD_LEN + 1];
    int state;
};

static char *enable_user_fast(int level, int try)
{
 if (level < TAC_PLUS_PRIV_LVL_MAX)
  {
   static char s[11];
   if(try)
     return NULL;
   sprintf(s, ENAB_PREFIX "enab%d" ENAB_PREFIX, level);
   return s;
  }

 return (try) ? ENAB_PREFIX "enab15" ENAB_PREFIX : ENAB_PREFIX "enable" ENAB_PREFIX;
}

static char *enable_user(int level, int try)
{
 if (level < TAC_PLUS_PRIV_LVL_MIN || level > TAC_PLUS_PRIV_LVL_MAX)
    return NULL;

 return enable_user_fast(level, try);
}

static void
enable(passwd, data)
char *passwd;
struct authen_data *data;
{
    int level = data->NAS_id->priv_lvl;

    char *name;
    int try;

#ifdef PERUSER_ENABLE
    char *cfg_enable;

    name = data->NAS_id->username;

    if(name && name[0] &&
      (cfg_enable = (char*)cfg_get_enable (name, TAC_PLUS_RECURSE)) != NULL)
     {
      if(!STREQ("login", cfg_enable))
        data->service = TAC_PLUS_AUTHEN_SVC_ENUSER;

      if(!verify(name, passwd, data, TAC_PLUS_RECURSE))
        data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;

      data->service = TAC_PLUS_AUTHEN_SVC_ENABLE;

      return;
     }

#endif

					/* standard 		*/

    for(try = 0; try < 2; try++)	/* two names possible	*/
     {
      if((name = enable_user_fast(level, try)) == NULL)
        break;				/* no name 		*/

      if(verify(name, passwd, data, TAC_PLUS_NORECURSE))
        return;
      }

    /* return fail */
    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;

    return;
}


/*
 * Tacacs enable authentication function. Wants an enable
 * password, and tries to verify it.
 *
 * Choose_authen will ensure that we already have a username before this
 * gets called.
 *
 * We will query for a password and keep it in the method_data.
 *
 * Any strings returned via pointers in authen_data must come from the
 * heap. They will get freed by the caller.
 *
 * Return 0 if data->status is valid, otherwise 1
 */

int
enable_fn(data)
struct authen_data *data;
{
    char *passwd;
    struct private_data *p;
    int pwlen;
    int level;
    char *name;

#ifdef NONE_LOGIN
    int try;
#endif
#if defined(PERUSER_ENABLE) || defined(NONE_LOGIN)
    char *cfg_enable;
    int nopassword = 0;
#endif

    p = (struct private_data *) data->method_data;

    /* An abort has been received. Clean up and return */
    if (data->flags & TAC_PLUS_CONTINUE_FLAG_ABORT) {
	if (data->method_data)
	    free(data->method_data);
	data->method_data = NULL;
	return (1);
    }
    /* Initialise method_data if first time through */
    if (!p) {
	p = (struct private_data *) tac_malloc(sizeof(struct private_data));
	bzero(p, sizeof(struct private_data));
	data->method_data = p;
	p->state = STATE_AUTHEN_START;
    }

    /* As we're enabling, we don't need a username, but do we have a
       password? */

    passwd = p->password;

    level = data->NAS_id->priv_lvl;
    name = data->NAS_id->username;

#if defined(PERUSER_ENABLE) || defined(NONE_LOGIN)
    while(1)
    {
#ifdef PERUSER_ENABLE
					/* per-user ? */
    if(name && name[0] &&
       (cfg_enable = (char*)cfg_get_enable(name, TAC_PLUS_RECURSE)) != NULL)
     {
      if(STREQ("permit", cfg_enable))
        nopassword = 1;
      else if(STREQ("deny", cfg_enable))
        nopassword = 2;
      else if(STREQ("none", cfg_enable))
        nopassword = -1;
      else if(STREQ("login", cfg_enable))
       {
        char *secret;
        if ((secret = cfg_get_login_secret(name, TAC_PLUS_RECURSE)) != NULL)
         {
          if(STREQ(secret, "permit"))
            nopassword = 1;
          else if(STREQ(secret, "deny"))
            nopassword = 2;
          else if(STREQ(secret, "none"))
            nopassword = -1;
         }
       }
      break;				/* per-user! - no next checks	*/
     }

#endif /* PERUSER_ENABLE */

#ifdef NONE_LOGIN

    for(try = 0; try < 2; try++)	/* two names possible	*/
     {
      if((name = enable_user(level, try)) == NULL)
        break;				/* no name 		*/

      if((cfg_enable = cfg_get_login_secret(name, TAC_PLUS_NORECURSE)) != NULL)
       {
        if(STREQ("permit", cfg_enable))
          nopassword = 1;
        else if(STREQ("deny", cfg_enable))
          nopassword = 2;
        else if(STREQ("none", cfg_enable))
          nopassword = -1;
        break;
       }
     }

#endif /* NONE_LOGIN */

    break;				/* no loops			*/
    }

    if(nopassword)
     ;
    else
#endif

    if (!passwd[0]) {

	/* No password. Either we need to ask for one and expect to get
	 * called again, or we asked but nothing came back, which is fatal */

	switch (p->state) {
	case STATE_AUTHEN_GETPASS:
	    /* We already asked for a password. This should be the
               reply */
	    if (data->client_msg) {
		pwlen = MIN((int)strlen(data->client_msg), MAX_PASSWD_LEN);
	    } else {
		pwlen = 0;
	    }
	    strncpy(passwd, data->client_msg, pwlen);
	    passwd[pwlen] = '\0';
	    break;

	default:
	    /* Request a password */
	    data->flags = TAC_PLUS_AUTHEN_FLAG_NOECHO;
	    fr_fi(data->server_msg, tac_strdup("Password: "));
	    data->status = TAC_PLUS_AUTHEN_STATUS_GETPASS;
	    p->state = STATE_AUTHEN_GETPASS;
	    return (0);
	}
    }

    /* We have a password. Try validating */

    /* Assume the worst */
    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;

    switch (data->service) {
    case TAC_PLUS_AUTHEN_SVC_ENABLE:
        /* sanity check */
        if (level < TAC_PLUS_PRIV_LVL_MIN || level > TAC_PLUS_PRIV_LVL_MAX) {
	  data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
	  fr_fi(data->server_msg, tac_strdup("Invalid privilege level in packet"));
	  report(LOG_ERR, "%s level=%d %s", session.peer, level, data->server_msg);
        }
        else
#if defined(PERUSER_ENABLE) || defined(NONE_LOGIN)
        if(nopassword > 0)	/* deny, permit - no procedure	*/
         {
          if(nopassword == 1)	/* permit - set flag */
            data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
         }
        else			/* none, password - do procedure */
#endif
	enable(passwd, data);
	if (debug || !(data->status == TAC_PLUS_AUTHEN_STATUS_PASS)) {
	    char *name = data->NAS_id->username;

	    report(LOG_INFO, "enable query for '%s' %s from %s %s",
		   name && name[0] ? name : "unknown",
		   data->NAS_id->NAS_port && data->NAS_id->NAS_port[0] ?
		       data->NAS_id->NAS_port : "unknown",
		   session.peer, 
		   (data->status == TAC_PLUS_AUTHEN_STATUS_PASS) ?
		   "accepted" : "rejected");
	}
#ifdef EMESSAGE
        if(data->status == TAC_PLUS_AUTHEN_STATUS_PASS && name && name[0] &&
           !data->server_msg)
	 {
	  char *msg;
	  if((msg = cfg_get_pvalue (name, TAC_IS_USER, S_emessage, TAC_PLUS_RECURSE)) != NULL)
	   {
	    data->server_msg =
#ifdef MORE_SUBST
                                       substitute2(msg, (void *)data);
#else
	                               tac_strdup(msg);
#endif
           }
	 }
#endif
	break;
    default:
	data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
	report(LOG_ERR, "%s: Bogus service value %d from packet", 
	       session.peer, data->service);
	break;
    }

    if (data->method_data)
	free(data->method_data);
    data->method_data = NULL;

    switch (data->status) {
    case TAC_PLUS_AUTHEN_STATUS_ERROR:
    case TAC_PLUS_AUTHEN_STATUS_FAIL:
    case TAC_PLUS_AUTHEN_STATUS_PASS:
	return (0);
    default:
	report(LOG_ERR, "%s: authenticate_fn can't set status %d",
	       session.peer, data->status);
	data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
	return (1);
    }
}
