/* 
   Copyright (c) 1995-2000 by Cisco systems, Inc.

   Permission to use, copy, modify, and distribute modified and
   unmodified copies of this software for any purpose and without fee is
   hereby granted, provided that (a) this copyright and permission notice
   appear on all copies of the software and supporting documentation, (b)
   the name of Cisco Systems, Inc. not be used in advertising or
   publicity pertaining to distribution of the program without specific
   prior permission, and (c) notice be given in supporting documentation
   that use, modification, copying and distribution is by permission of
   Cisco Systems, Inc.

   Cisco Systems, Inc. makes no representations about the suitability
   of this software for any purpose.  THIS SOFTWARE IS PROVIDED ``AS
   IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
   WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
   FITNESS FOR A PARTICULAR PURPOSE.
*/


#include <ctype.h>
#include <time.h>

#include "tac_plus.h"
#include "expire.h"

/*
 * check a date for expiry. If the field specifies
 * a shell return PW_OK
 *
 * Return PW_OK if not expired
 * Return PW_EXPIRING if expiry is coming soon
 * Return PW_EXPIRED  if already expired
 */

#define SEC_IN_DAY (24*60*60)

#ifdef WARNEXP
 #ifdef FINEEXP
   #define WARNING_PERIOD WARNEXP
 #else
   #define WARNING_PERIOD (WARNEXP / SEC_IN_DAY * SEC_IN_DAY)
 #endif
#else
 #define WARNING_PERIOD (14 * SEC_IN_DAY)
#endif

static char *monthname[] = {"JAN", "FEB", "MAR", "APR", "MAY", "JUN",
"JUL", "AUG", "SEP", "OCT", "NOV", "DEC"};

#ifndef LOCEXP
static long days_ere_month[] = {0, 31, 59, 90, 120, 151,
181, 212, 243, 273, 304, 334};
#endif

#ifndef FINEEXP

static time_t align_time(time_t t)
{
#ifdef LOCEXP
    struct tm a, *b;

    b = localtime(&t);

    a.tm_sec = 0;
    a.tm_min = 0;
    a.tm_hour = 0;
    a.tm_mday = b->tm_mday;
    a.tm_mon = b->tm_mon;
    a.tm_year = b->tm_year;
    a.tm_isdst = -1;
#ifdef HAVE_STRUCT_TM_TM_ZONE
    a.tm_zone = 0;
    a.tm_gmtoff = 0;
#endif

    return mktime(&a);
#else
    return t / SEC_IN_DAY * SEC_IN_DAY;
#endif
}

#endif


int
check_expiration(date)
char *date;
{
    long day, month, year, now, expiration, warning;
    char monthstr[10];
    int i;
    char *s;

    monthstr[0] = '\0';

    /* If no date or a shell, let it pass.  (Backward compatibility.) */
    if (!date || (strlen(date) == 0) || (*date == '/'))
	return (PW_OK);

    /* Skip lead spaces */
    while(*date != '\0' && (*date == ' ' || *date == '\t'))
	date++;

    /* Is it all? No date, let it pass */
    if (*date == '\0')
	return (PW_OK);

    /* Skip digits */
    for(s = date; *s != '\0' && isdigit((int)*s); s++)
      ;

    /* Detect 1st non-digit/trailing space character */
    while(*s != '\0' && (*s == ' ' || *s == '\t'))
	s++;

    if(*s == '\0')		/* date in system format */
    {
      expiration = 	/* convert to int, GMT	*/
#ifndef FINEEXP
                    align_time(
#endif
                               atol(date)
#ifndef FINEEXP
                              )
#endif
                               ;
    } else {			/* date in tac+ format */
#ifdef LOCEXP
	struct tm t;
#endif

	/* Parse date string.  Fail it upon error. */
	if (sscanf(date, "%s %ld %ld", monthstr, &day, &year) != 3)
	    return (PW_EXPIRED);

	for(i=0; i < 3; i++) {
	    monthstr[i] = toupper(monthstr[i]);
	}

	/* Compute the expiration date in days. */
	for (month = 0; month < 12; month++)
	    if (strncmp(monthstr, monthname[month], 3) == 0)
  		break;

	if (month > 11)
	    return (PW_EXPIRED);

#ifdef LOCEXP
	t.tm_sec = 0;
	t.tm_min = 0;
	t.tm_hour = 0;
	t.tm_mday = day;
	t.tm_mon = month;
	t.tm_year = year - 1900;
	t.tm_isdst = -1;
#ifdef HAVE_STRUCT_TM_TM_ZONE
	t.tm_zone = 0;
	t.tm_gmtoff = 0;
#endif

	expiration = mktime(&t);	/* int, GMT, aligned	*/
#else
	expiration = (
                    ((year - 1970) * 365) + days_ere_month[month] + (day - 1) +
                    ((year - 1969) / 4 + (((year % 4) == 0) && (month > 2)))
                   ) * SEC_IN_DAY;
#endif
    }


    /* Get the current time , int, GMT */
    now =
#ifndef FINEEXP
                    align_time(
#endif
                               time(NULL)
#ifndef FINEEXP
                              )
#endif
                               ;

    warning = difftime(expiration, now);

    if(warning
#ifdef FINEEXP
               <=
#else
               <
#endif
                 0)
	return (PW_EXPIRED);

    if(warning
#ifdef FINEEXP
               <=
#else
               <
#endif
                 WARNING_PERIOD)
	return (PW_EXPIRING);

    return (PW_OK);
}
